# Imports from stdlib:
import time
import socketserver
import struct
import threading
import logging

# Internal imports:
from . import modbus_datastore

# Development imports:
import stackprinter
from icecream import ic

socketserver.TCPServer.allow_reuse_address = True
logging.basicConfig(format="%(levelname)s: %(message)s", level=logging.DEBUG)


def pack_bools_to_bytes(bool_list):
    chars = []
    for j in range(0, len(bool_list), 8):
        bool_chunk = bool_list[j : j + 8]
        equivalent_char = 0
        for i, b in enumerate(bool_chunk):
            if b:
                equivalent_char += 2 ** i
        chars.append(equivalent_char)

    return struct.pack(f"!{len(chars)}B", *chars)


def build_error_response(header_items, exception_code):
    response_items = list(header_items)

    # Error Response Function Code -> Function Code + 128
    response_items[4] = header_items[4] + 128

    # In an error message, the data contains the Exception Code
    response_items.append(exception_code)

    response = struct.pack(f"!HHHBBB", *response_items)

    error_descriptions = {1: "Illegal Function code", 2: "", 3: "", 4: ""}
    logging.info(f"Modbus Error {exception_code}")

    return response


class TCPHandler(socketserver.BaseRequestHandler):
    def handle(self):
        """Callback Function for requests to the TCP Server:"""

        # Recv max 255 bytes, the maximal length for a Modbus frame:
        self.data = self.request.recv(255).strip()

        ## Extract Header + Function Code:
        # Transaction ID:   (2 Bytes)   Identifies the request-response-pair, is echoed in the response
        # Protocol:         (2 Bytes)   Always 0 ("reserved for future use", lol)
        # Length:           (2 Bytes)   Length of the remaining frame in bytes (Total Length - 6)
        # Unit ID:          (1 Byte)    "Slave ID", inner identifier to route to different units (typically 0)
        # Function Code:    (1 Byte)    1,2,3,4,5,6,15,16,43: Read/Write input/register etc.
        transaction_id, protocol, length, unit_id, function_code = struct.unpack(
            "!HHHBB", self.data[:8]
        )

        # Pack header items into a tuple which can more easily be passed around:
        header_items = (transaction_id, protocol, length, unit_id, function_code)

        # Check if Function Code is valid:
        if function_code not in (1, 2, 3, 4):
            # Respond with exception 01 - Illegal Function:
            response = build_error_response(header_items, exception_code=1)
            self.request.sendall(response)
            return None

        if function_code in (1, 2, 3, 4):  # 'Read' Function Codes
            first_address = struct.unpack("!H", self.data[8:10])[0]
            number_of_registers = struct.unpack("!H", self.data[10:12])[0]

        function_code_map = {
            1: "coils",
            2: "discrete_inputs",
            4: "input_registers",
            3: "holding_registers",
        }
        object_reference = function_code_map[function_code]

        ## Validate number of objects requested and respond with exception 3 if invalid:
        ## =============================================================================

        if object_reference in ("coils", "discrete_inputs"):
            if number_of_registers < 1 or number_of_registers > 2000:
                response = build_error_response(header_items, exception_code=3)
                self.request.sendall(response)
                return None

        if object_reference in ("input_registers", "holding_registers"):
            if number_of_registers < 1 or number_of_registers > 125:
                response = build_error_response(header_items, exception_code=3)
                self.request.sendall(response)
                return None

        ## Read addresses from datastore
        ## =============================

        try:
            # self.server is a reference to the socketserver.ThreadingTCPServer instance defined below:
            data = self.server.datastore.read(
                object_reference, first_address, number_of_registers
            )
        except KeyError:
            # Address not in datastore -> Respond with exception 02 - Illegal Data Address:
            response = build_error_response(header_items, exception_code=2)
            self.request.sendall(response)
            return
        except:
            raise
            # Other Error -> Respond with exception 04 - Slave Device Failure:
            response = build_error_response(header_items, exception_code=4)
            self.request.sendall(response)
            return

        ## Compose response
        ## ================

        if object_reference in ("coils", "discrete_inputs"):
            data_bytes = pack_bools_to_bytes(data)

        if object_reference in ("input_registers", "holding_registers"):
            data_bytes = b"".join(data)

        # Response length is 3 fixed bytes (unit_id, function_code, number_of_data_bytes) plus the data bytes:
        response_message_length = 3 + len(data_bytes)
        number_of_data_bytes = len(data_bytes)

        # Compose response header
        response_items = [
            transaction_id,
            protocol,
            response_message_length,
            unit_id,
            function_code,
            number_of_data_bytes,
        ]

        # Pack response items into binary format and append data_bytes:
        response = struct.pack(f"!HHHBBB", *response_items) + data_bytes

        self.request.sendall(response)


class Server:
    def __init__(
        self,
        host="localhost",
        port=502,
        daemon=False,
        datastore=modbus_datastore.DictDatastore(),
    ):
        self.host = host
        self.port = port
        self.datastore = datastore
        self.daemon = daemon
        self.tcp_server = None

    def _server_thread(self):
        self.tcp_server = socketserver.ThreadingTCPServer(
            (self.host, self.port), TCPHandler
        )
        self.tcp_server.allow_reuse_address = True
        self.tcp_server.datastore = self.datastore
        self.tcp_server.serve_forever()

    def start(self):
        self.server_thread = threading.Thread(target=self._server_thread)
        # daemon == True -> Shutdown the thread when the main program exits
        self.server_thread.daemon = self.daemon
        self.server_thread.start()

    def stop(self):
        if self.tcp_server:
            self.tcp_server.shutdown()
            self.tcp_server.server_close()
            self.tcp_server = None
        logging.info("Modbus Server stopped")

    ## Convenience Functions for direct access to object reference (single + multiple):
    ## ================================================================================

    def set_coil(self, address, value):
        self._set_value("coils", address, value)

    def set_coils(self, address, values):
        for value in values:
            self._set_value("coils", address, value)
            address += 1

    def set_discrete_input(self, address, value):
        self._set_value("discrete_inputs", address, value)

    def set_discrete_inputs(self, address, values):
        for value in values:
            self._set_value("discrete_inputs", address, value)
            address += 1

    def set_input_register(self, address, value, encoding):
        self._set_value("input_registers", address, value, encoding)

    def set_input_registers(self, address, values, encoding):
        for value in values:
            self._set_value("input_registers", address, value, encoding)
            address += struct.calcsize(encoding) // 2

    def set_holding_register(self, address, value, encoding):
        self._set_value("holding_registers", address, value, encoding)

    def set_holding_registers(self, address, values, encoding):
        for value in values:
            self._set_value("holding_registers", address, value, encoding)
            address += struct.calcsize(encoding) // 2

    # Actually set the value:
    # =======================

    def _set_value(self, object_reference, address, value, encoding=None):

        # Verify address:
        if type(address) is not int:
            raise TypeError(f"type of 'address' must be int, not {type(address)}")
        if address < 0 or address > 65535:
            raise ValueError(f"'address' must be between 0 and 65535, not {address}")

        # Verify if value is boolean for coils and discrete inputs:
        # This is not done with bool(), because bool('0') == True might be confusing
        if object_reference in ("coils", "discrete_inputs") and not type(value) == bool:
            raise TypeError(
                f"'value' for {object_reference} must be True or False, is {type(value)}"
            )

        # Verify if value can be converted to float for input_registers and holding_registers:
        # This works for float, int, and string with valid number inside
        if object_reference in ("input_registers", "holding_registers"):

            # Verify encoding:
            if encoding not in ("h", "H", "e"):
                raise ValueError(
                    f'encoding must be "h"(short), "H"(unsigned short), or "e"(float16), not {encoding}'
                )

        self.datastore.write(object_reference, address, value, encoding)



import json
import simple_modbus_server as sms

with open("example_modbus_address_map.json") as f:
    modbus_address_map = json.load(f)

datastore = sms.RedisDatastore(modbus_address_map=modbus_address_map)

modbus_server = sms.Server(port=5020, datastore=datastore)

modbus_server.start()

modbus_server.set_coil(0, True)
modbus_server.set_coil(1, False)



import simple_modbus_server as sms

server = sms.Server(host="localhost", port=5020)

server.set_value(object_reference=3, address=1, value=19)

server.start()
print("Started Modbus Server...")

# With Redis Datastore:
# =====================

with open("modbus_address_map") as f:
    modbus_address_map = json.load(f)

datastore = SimpleModbusServer.RedisDatastore(
    host="localhost", port=6379, db=0, modbus_address_map=modbus_address_map
)
sms.Server(host="localhost", port=5020, datastore=datastore)


# For Registers:
# ==============
# Encoding can be h (short), H (unsigned short), e (float16)
server.set_input_register(0, 1.5, encoding="e")



import struct
import importlib.util
import distutils.util
import json
import warnings

import redis


class DictDatastore:
    def __init__(self):
        self.datadict = {
            "coils": {},
            "discrete_inputs": {},
            "input_registers": {},
            "holding_registers": {},
        }

    def read(self, object_reference, first_address, quantity_of_records):
        data = []
        for address in range(first_address, first_address + quantity_of_records):
            data.append(self.datadict[object_reference][address])
        return data

    def write(self, object_reference, address, value, encoding):

        if object_reference in ("input_registers", "holding_registers"):
            if struct.calcsize(encoding) == 2:
                value = struct.pack(f"!{encoding}", value)
            elif struct.calcsize(encoding) > 2:
                number_of_registers = struct.calcsize(encoding) // 2
                value_bytes = struct.pack(f"!{encoding}", value)
                for chunk_number in range(0, number_of_registers):
                    byte_chunk = value_bytes[chunk_number * 2, chunk_number * 2 + 2]
                    self.datadict[object_reference][address + chunk_number] = value
                return

        self.datadict[object_reference][address] = value

    def dump(self):
        return self.datadict


class RedisDatastore:
    def __init__(self, host="localhost", port=6379, db=0, modbus_address_map={}):
        if importlib.util.find_spec("redis") is not None:
            import redis
        else:
            raise ImportError(
                "RedisDatastore requires redis-py (https://github.com/andymccurdy/redis-py),\
                which you can install with 'pip install redis'"
            )
        self.host = host
        self.port = port
        self.db = db
        self.modbus_address_map = modbus_address_map
        self.r = None
        self._verify_modbus_address_map()
        self._connect()

    def _connect(self):
        self.r = redis.Redis(self.host, self.port, self.db)

    def _verify_modbus_address_map(self):
        for key in self.modbus_address_map.keys():
            if key not in ("coils", "discrete_inputs", "input_registers", "holding_registers"):
                warnings.warn(f"modbus_address_map contains non-standard key {key}")
        for std_key in ("coils", "discrete_inputs", "input_registers", "holding_registers"):
            if std_key not in self.modbus_address_map:
                self.modbus_address_map[std_key] = {}

    def read(self, object_reference, first_address, quantity_of_records):
        data = []
        for address in range(first_address, first_address + quantity_of_records):
            key = self.modbus_address_map[object_reference][address]["key"]
            raw_value = self.r.get(key)

            if object_reference in ("input_registers", "holding_registers"):
                encoding = self.modbus_address_map[object_reference][address]["encoding"]

                if encoding in ("h", "H", "i"):  # ints
                    cast = int
                if encoding in ("e", "f", "d"):  # floats
                    cast = float
                value = struct.pack(f"!{encoding}", cast(raw_value))

                if encoding in ("i", "I", "f", "d"):
                    # >2 bytes: Find out which part is requested:
                    part = self.modbus_address_map[object_reference][address]["part"]
                    value_bytes = struct.pack(f"!{encoding}", value)
                    value = value_bytes[part * 2 : part * 2 + 2]

            elif object_reference in ("coils", "discrete_inputs"):
                # Cast value from string to bool:
                value = bool(distutils.util.strtobool(raw_value.decode()))

            data.append(value)

        return data

    def write(self, object_reference, address, value, encoding):
        if type(value) == bool:
            value = str(value)
        try:
            key = self.modbus_address_map[object_reference][address]["key"]
        except KeyError:
            key = f"{object_reference}:{address}"
            self.modbus_address_map[object_reference][address] = {
                "key": key,
                "encoding": encoding,
            }
        self.r.set(key, value)



import time

import chalk
from icecream import ic

from pymodbus.client.sync import ModbusTcpClient

import modbus_server

# Helper Function to print yellow with yellow()
yellow = lambda s: print(chalk.yellow(s))


class pymodbus_client:
    def __init__(self, host="127.0.0.1", port=5020):
        self.client = ModbusTcpClient(host=host, port=port)

    def __enter__(self):
        return self.client

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.client.close()


# def test_illegal_function():
#     with pymodbus_client() as client:
#         result = client.write_coil(address=0, value=True)
#         # pprint(dir(result))
#         print(result.fcode)
#         print(result.message)
#         assert result.isError() == True


def assert_bits(obj_ref, address, count=1, expect_result=True, expect_error=False):
    with pymodbus_client() as client:
        if obj_ref == "coils":
            result = client.read_coils(address=address, count=count)
        elif obj_ref == "discrete_inputs":
            result = client.read_discrete_inputs(address=address, count=count)
        else:
            raise ValueError()
        assert result.isError() == expect_error
        if not expect_error:
            if type(expect_result) == bool:
                assert result.bits[:count] == [expect_result] * count
            elif type(expect_result) in (list, tuple):
                assert result.bits == expect_result


def assert_registers(obj_ref, address, count=1, expect_result=0, expect_error=False):
    with pymodbus_client() as client:
        if obj_ref == "input_registers":
            result = client.read_input_registers(address=address, count=count)
        elif obj_ref == "holding_registers":
            result = client.read_holding_registers(address=address, count=count)
        else:
            raise ValueError()
        assert result.isError() == expect_error
        if not expect_error:
            ic(result.registers)
            if type(expect_result) in (int, float):
                assert result.registers[:count] == [expect_result] * count
            elif type(expect_result) in (list, tuple):
                assert result.registers == expect_result


def check_bits(s, obj_ref):

    s.set_coil(0, True)
    s.set_coil(1, False)
    s.set_coils(10000, values=[True] * 200)

    s.set_discrete_input(0, True)
    s.set_discrete_input(1, False)
    s.set_discrete_inputs(10000, values=[True] * 200)

    # Read Addresses 0 and 1:
    assert_bits(obj_ref, address=0, expect_result=True)
    assert_bits(obj_ref, address=1, expect_result=False)

    # Read Addresses 10000 - 10200:
    assert_bits(obj_ref, address=10000, count=200)

    # Expect Exception 2 - Illegal Data Address:
    assert_bits(obj_ref, address=0, count=3, expect_error=True)
    assert_bits(obj_ref, address=3, count=1, expect_error=True)

    # Expect Exception 3 - Illegal Data Value:
    assert_bits(obj_ref, address=0, count=0, expect_error=True)
    assert_bits(obj_ref, address=0, count=2000, expect_error=True)


def check_registers(s, obj_ref):

    s.set_input_register(0, 13, encoding="h")
    s.set_input_register(1, 14, encoding="h")
    s.set_input_registers(10000, values=[1000, 1001, 1002], encoding="H")

    s.set_holding_register(0, 15, encoding="h")
    s.set_holding_register(1, 16, encoding="h")
    s.set_holding_registers(10000, values=[1000, 1001, 1002], encoding="H")

    # Read Addresses 0 and 1:
    assert_registers(obj_ref, address=0, expect_result=True)
    assert_registers(obj_ref, address=1, expect_result=False)

    # Read Addresses 10000 - 10200:
    assert_registers(obj_ref, address=10000, count=3, expect_result=[1000, 1001, 1002])

    # Expect Exception 2 - Illegal Data Address:
    assert_registers(obj_ref, address=0, count=3, expect_error=True)
    assert_registers(obj_ref, address=3, count=1, expect_error=True)

    # Expect Exception 3 - Illegal Data Value:
    assert_registers(obj_ref, address=0, count=0, expect_error=True)
    assert_registers(obj_ref, address=0, count=126, expect_error=True)


def test_bits_with_DictDatastore():
    s = modbus_server.Server(host="localhost", port=5020, daemon=True)
    s.start()
    time.sleep(0.5)

    check_bits(s, "coils")
    check_bits(s, "discrete_inputs")

    s.stop()


def test_bits_with_RedisDatastore():
    datastore = modbus_server.RedisDatastore()
    s = modbus_server.Server(
        host="localhost", port=5020, daemon=True, datastore=datastore
    )
    s.start()
    time.sleep(0.5)

    check_bits(s, "coils")
    check_bits(s, "discrete_inputs")

    s.stop()


def test_registers_with_DictDatastore():
    s = modbus_server.Server(host="localhost", port=5020, daemon=True)
    s.start()
    time.sleep(0.5)

    check_registers(s, "input_registers")
    check_registers(s, "holding_registers")

    s.stop()


def test_registers_with_RedisDatastore():
    datastore = modbus_server.RedisDatastore()
    s = modbus_server.Server(
        host="localhost", port=5020, daemon=True, datastore=datastore
    )
    s.start()
    time.sleep(0.5)

    check_registers(s, "input_registers")
    check_registers(s, "holding_registers")

    s.stop()


def now_test_discrete_inputs():
    s = modbus_server.Server(host="localhost", port=5020)
    s.start()
    time.sleep(0.5)

    s.set_discrete_input(0, value=False)
    s.set_discrete_input(1, value=True)

    print(chalk.yellow("Request address normally:"))
    with pymodbus_client() as client:
        result = client.read_discrete_inputs(address=0, count=2)
        assert result.isError() == False
        assert result.bits[0] == False
        assert result.bits[1] == True
        print(result.bits[2:8])
        assert result.bits[2:8] == [False] * 6

    s.set_discrete_inputs(10000, values=[True] * 200)
    with pymodbus_client() as client:
        result = client.read_discrete_inputs(address=10000, count=200)
        assert result.isError() == False
        assert result.bits == [True] * 200

    # Exception 2 - Illegal Data Address
    print(chalk.yellow("Request address that is not set:"))
    with pymodbus_client() as client:
        result = client.read_discrete_inputs(address=0, count=3)
        assert result.isError() == True

    s.stop()


def now_test_read_input_registers():

    ## Test FC 3
    ## =========

    s.set_input_register(0, value=14, encoding="h")
    s.set_input_register(1, value=15, encoding="h")

    print(chalk.yellow("Request address normally:"))
    with pymodbus_client() as client:
        result = client.read_input_registers(address=0, count=2)
        assert result.isError() == False
        assert result.registers[0] == 14
        assert result.registers[1] == 15

    # s.set_discrete_inputs(10000, values=[True] * 200)
    # with pymodbus_client() as client:
    #     result = client.read_discrete_inputs(address=10000, count=200)
    #     assert result.isError() == False
    #     assert result.bits == [True] * 200

    # Exception 2 - Illegal Data Address
    print(chalk.yellow("Request address that is not set:"))
    with pymodbus_client() as client:
        result = client.read_input_registers(address=0, count=3)
        assert result.isError() == True


if __name__ == "__main__":
    from pprint import pprint
    import inspect

    # test_functions = [k for k, v in globals().items() if inspect.isfunction(v)]

    # for func in test_functions:
    #     globals()[func]()

    # Init Modbus Server:
    datastore = modbus_server.RedisDatastore()
    s = modbus_server.Server(
        host="localhost", port=5020, daemon=True, datastore=datastore
    )
    s.start()
    time.sleep(0.5)

    # now_test_read_input_registers()
    test_bits_with_RedisDatastore()

    s.stop()



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['modbus_server']

package_data = \
{'': ['*']}

setup(name='modbus_server',
      version='0.1.0',
      description='Easy-to-use ModbusTCP Server in pure Python ',
      author='Christoph Kattmann',
      author_email='christoph.kattmann@bss-hs.de',
      url=None,
      packages=packages,
      package_data=package_data,
     )
